; ------------------------------------------------------------
; File "MLsort.pbi" 
; Merge List Sort (in-place) with variable comparison function 
; sort command: LLmergeSort(@CompareProcedure()) 
; by horst.schaeffer@gmx.net 
; ------------------------------------------------------------
Macro Until_Break : ForEver : EndMacro   ; nothing runs forever ;)

; Getting addres of next element (independent of PB's "current element")
; note: the next pointer of last list element is NUL

#LLnextOffset = 2*SizeOf(integer)  ; pointer at negative offset
Macro SetNextElement(pointer)
  pointer - #LLnextOffset
  pointer = pointer\i + #LLnextOffset  
EndMacro 
#EndOfList = #LLnextOffset  ; at end of list pointer will be value of offset after macro 

Prototype MLcmp(*first,*second)   ; for user defined comparison isInRightOrder()
; UserList() must be assigned BEFORE (by macro) to an already defined list
; ++++++++++++++++++++++++++++++++++++++++++++++++++++
; Merging:
; FirstList, SecondList are adjacent parts of the linked list.
; Merging done by moving elements from second to first list.

Procedure MergeSections(*FirstList.integer,FirstListSize,*SecondList.integer,SecondListSize) 
  Protected loop, *MergedList
  Shared isInRightOrder.MLcmp                           ; user defined comparison 
  
  *MergedList = *FirstList                              ; to return, may be corrected *
  Repeat 
    If isInRightOrder(*FirstList,*SecondList)           ; elements in right (user defined) order? 
      If FirstListSize = 1 : Break : EndIf              ; done, *SecondList is succeeding
      SetNextElement(*FirstList)
      FirstListSize -1
    Else 
      ChangeCurrentElement(UserList(),*SecondList)       ; this one to move
      SetNextElement(*SecondList)                        ; element to continue with after move
      MoveElement(UserList(),#PB_List_Before,*FirstList)
      If loop = 0 : *MergedList = @UserList() : EndIf    ; entry pointer is the one just moved * 
      If SecondListSize = 1 : Break : EndIf              ; and that was the one just moved!
      SecondListSize -1
    EndIf 
    loop +1
  Until_Break
ProcedureReturn *MergedList
EndProcedure

; ++++++++++++++++++++++++++++++++++++++++++++++++++++

Procedure RecursiveChopAndMerge(*TargetList,TargetListSize)  
  Shared *RestOfList.integer, TotalSortSize    ; shared with LLmergeSort()
  Shared *SourceList.integer, SourceListSize   ; shared with other recursion instances 
  
  *SourceList.integer = *RestofList    ; chop off another element from rest of list
  SetNextElement(*RestOfList)
  SourceListSize = 1
  
  While SourceListSize < TargetListSize And *RestOfList <> #EndOfList
    RecursiveChopAndMerge(*SourceList,SourceListSize)     ; produces new source
  Wend 
  
  If TargetListSize <> TotalSortSize
    *SourceList = MergeSections(*TargetList,TargetListSize,*SourceList,SourceListSize)
    SourceListSize = TargetListSize + SourceListSize
  EndIf 
EndProcedure

; ++++++++++++++++++++++++++++++++++++++++++++++++++++

Procedure LLmergeSort(ComparisonProcedure,SkipLines=0) 
  Shared isInRightOrder.MLcmp, *RestofList, TotalSortSize
  isInRightOrder = ComparisonProcedure
  TotalSortSize = ListSize(UserList()) - SkipLines 
  If TotalSortSize > 1
    *RestofList = SelectElement(UserList(),SkipLines)
    RecursiveChopAndMerge(#NUL,TotalSortSize)
  EndIf 
EndProcedure  

; ++++++++++++++++++++++++++++++++++++++++++++++++++++



; IDE Options = PureBasic 4.60 Beta 4 (Windows - x86)
; CursorPosition = 79
; FirstLine = 43
; Folding = -
; CompileSourceDirectory